(function(exports) {
    "use strict";

    exports.create = function create(options) {
        return new this.ClientUtils(options);
    };

    exports.ClientUtils = function ClientUtils(options) {
        var BASE64_ENCODE_CHARS = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
        var BASE64_DECODE_CHARS = [
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 62, -1, -1, -1, 63,
            52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1, -1, -1, -1, -1, -1,
            -1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, -1,
            -1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
            41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, -1, -1, -1, -1, -1
        ];
        var SUPPORTED_SELECTOR_TYPES = ['css', 'xpath'];
        var XPATH_NAMESPACE = {
            svg: 'http://www.w3.org/2000/svg',
            mathml: 'http://www.w3.org/1998/Math/MathML'
        };

        function form_urlencoded(str) {
            return encodeURIComponent(str)
                    .replace(/%20/g, '+')
                    .replace(/[!'()*]/g, function(c) {
                        return '%' + c.charCodeAt(0).toString(16);
                    });
        }

        this.options = options || {};
        this.options.scope = this.options.scope || document;

        this.__call = function __call(method, args) {
            if (method === "__call") {
                return;
            }
            try {
                return this[method].apply(this, args);
            } catch (err) {
                err.__isCallError = true;
                return err;
            }
        };

        this.click = function click(selector, x, y) {
            return this.mouseEvent('click', selector, x, y);
        };

        this.decode = function decode(str) {
            var c1, c2, c3, c4, i = 0, len = str.length, out = "";
            while (i < len) {
                do {
                    c1 = BASE64_DECODE_CHARS[str.charCodeAt(i++) & 0xff];
                } while (i < len && c1 === -1);
                if (c1 === -1) {
                    break;
                }
                do {
                    c2 = BASE64_DECODE_CHARS[str.charCodeAt(i++) & 0xff];
                } while (i < len && c2 === -1);
                if (c2 === -1) {
                    break;
                }
                out += String.fromCharCode(c1 << 2 | (c2 & 0x30) >> 4);
                do {
                    c3 = str.charCodeAt(i++) & 0xff;
                    if (c3 === 61) {
                        return out;
                    }
                    c3 = BASE64_DECODE_CHARS[c3];
                } while (i < len && c3 === -1);
                if (c3 === -1) {
                    break;
                }
                out += String.fromCharCode((c2 & 0XF) << 4 | (c3 & 0x3C) >> 2);
                do {
                    c4 = str.charCodeAt(i++) & 0xff;
                    if (c4 === 61) {
                        return out;
                    }
                    c4 = BASE64_DECODE_CHARS[c4];
                } while (i < len && c4 === -1);
                if (c4 === -1) {
                    break;
                }
                out += String.fromCharCode((c3 & 0x03) << 6 | c4);
            }
            return out;
        };

        this.echo = function echo(message) {
            console.log("[casper.echo] " + message);
        };
        this.elementVisible = function elementVisible(elem) {
        var style;
        try {
            style = window.getComputedStyle(elem, null);
        } catch (e) {
            return false;
        }
            if(style.visibility === 'hidden' || style.display === 'none') return false;
            var cr = elem.getBoundingClientRect();
            return cr.width > 0 && cr.height > 0;
        };

        this.encode = function encode(str) {
            var out = "", i = 0, len = str.length, c1, c2, c3;
            while (i < len) {
                c1 = str.charCodeAt(i++) & 0xff;
                if (i === len) {
                    out += BASE64_ENCODE_CHARS.charAt(c1 >> 2);
                    out += BASE64_ENCODE_CHARS.charAt((c1 & 0x3) << 4);
                    out += "==";
                    break;
                }
                c2 = str.charCodeAt(i++);
                if (i === len) {
                    out += BASE64_ENCODE_CHARS.charAt(c1 >> 2);
                    out += BASE64_ENCODE_CHARS.charAt((c1 & 0x3) << 4 | (c2 & 0xF0) >> 4);
                    out += BASE64_ENCODE_CHARS.charAt((c2 & 0xF) << 2);
                    out += "=";
                    break;
                }
                c3 = str.charCodeAt(i++);
                out += BASE64_ENCODE_CHARS.charAt(c1 >> 2);
                out += BASE64_ENCODE_CHARS.charAt((c1 & 0x3) << 4 | (c2 & 0xF0) >> 4);
                out += BASE64_ENCODE_CHARS.charAt((c2 & 0xF) << 2 | (c3 & 0xC0) >> 6);
                out += BASE64_ENCODE_CHARS.charAt(c3 & 0x3F);
            }
            return out;
        };

        this.exists = function exists(selector) {
            try {
                return this.findAll(selector).length > 0;
            } catch (e) {
                return false;
            }
        };

        this.fetchText = function fetchText(selector) {
            var text = '', elements = this.findAll(selector);
            if (elements && elements.length) {
                Array.prototype.forEach.call(elements, function _forEach(element) {
                    text += element.textContent || element.innerText || element.value || '';
                });
            }
            return text;
        };

        this.fill = function fill(form, vals, findType) {
            findType = findType || "names";

            var out = {
                errors: [],
                fields: [],
                files: []
            };

            if (!(form instanceof HTMLElement) || typeof form === "string") {
                this.log("attempting to fetch form element from selector: '" + form + "'", "info");
                try {
                    form = this.findOne(form);
                } catch (e) {
                    if (e.name === "SYNTAX_ERR") {
                        out.errors.push("invalid form selector provided: '" + form + "'");
                        return out;
                    }
                }
            }

            if (!form) {
                out.errors.push("form not found");
                return out;
            }

            for (var fieldSelector in vals) {
                if (!vals.hasOwnProperty(fieldSelector)) {
                    continue;
                }
                try {
                    out.fields[fieldSelector] = this.setFieldValue(this.makeSelector(fieldSelector, findType), vals[fieldSelector], form);
                } catch (err) {
                    switch (err.name) {
                        case "FieldNotFound":
                            out.errors.push('Unable to find field element in form: ' + err.toString());
                        break;
                        case "FileUploadError":
                            out.files.push({
                                type: findType,
                                selector: findType === "labels" ? '#' + err.id : fieldSelector,
                                path: err.path
                            });
                        break;
                        default:
                            out.errors.push(err.toString());
                    }
                }
            }
            return out;
        };
        this.findAll = function findAll(selector, scope) {
            scope = scope instanceof HTMLElement ? scope : scope && this.findOne(scope) || this.options.scope;
            try {
                var pSelector = this.processSelector(selector);
                if (pSelector.type === 'xpath') {
                    return this.getElementsByXPath(pSelector.path, scope);
                } else {
                    return Array.prototype.slice.call(scope.querySelectorAll(pSelector.path));
                }
            } catch (e) {
                this.log('findAll(): invalid selector provided "' + selector + '":' + e, "error");
            }
        };

        this.findOne = function findOne(selector, scope) {
            scope = scope instanceof HTMLElement ? scope : scope && this.findOne(scope) || this.options.scope;
            try {
                var pSelector = this.processSelector(selector);
                if (pSelector.type === 'xpath') {
                    return this.getElementByXPath(pSelector.path, scope);
                } else {
                    return scope.querySelector(pSelector.path);
                }
            } catch (e) {
                this.log('findOne(): invalid selector provided "' + selector + '":' + e, "error");
            }
        };

        this.forceTarget = function forceTarget(selector, target) {
            var elem = this.findOne(selector);
            while (!!elem && elem.tagName !== 'A' &&  elem.tagName !== 'FORM' && elem.tagName !== 'BODY'){
                elem = elem.parentNode;
            }
            if (elem === 'A' || elem === 'FORM') {
                elem.setAttribute('target', target);
                return true;
            }
            return false;
        };

        this.getBase64 = function getBase64(url, method, data) {
            return this.encode(this.getBinary(url, method, data));
        };
        this.getBinary = function getBinary(url, method, data) {
            try {
                return this.sendAJAX(url, method, data, false, {
                    overrideMimeType: "text/plain; charset=x-user-defined"
                });
            } catch (e) {
                if (e.name === "NETWORK_ERR" && e.code === 101) {
                    this.log("getBinary(): Unfortunately, casperjs cannot make"
                        + " cross domain ajax requests", "warning");
                }
                this.log("getBinary(): Error while fetching " + url + ": " + e, "error");
                return "";
            }
        };


        this.getCssSelector = function getCssSelector(selector, scope, limit) {
            scope = scope || this.options.scope;
            limit = limit || 'BODY';
            var elem = (selector instanceof Node) ? selector : this.findOne(selector, scope);
            if (!!elem && elem.nodeName !== "#document") {
                var str = "";
                while (elem.nodeName.toUpperCase() !== limit.toUpperCase()) {
                    str = "> " + elem.nodeName + ':nth-child(' + ([].indexOf.call(elem.parentNode.children, elem) + 1) + ') ' + str;
                    elem = elem.parentNode;
                }
                return str.substring(2);
            }
            return "";
        };

        this.getDocumentHeight = function getDocumentHeight() {
            return Math.max(
                Math.max(document.body.scrollHeight, document.documentElement.scrollHeight),
                Math.max(document.body.offsetHeight, document.documentElement.offsetHeight),
                Math.max(document.body.clientHeight, document.documentElement.clientHeight)
            );
        };

        this.getDocumentWidth = function getDocumentWidth() {
            return Math.max(
                Math.max(document.body.scrollWidth, document.documentElement.scrollWidth),
                Math.max(document.body.offsetWidth, document.documentElement.offsetWidth),
                Math.max(document.body.clientWidth, document.documentElement.clientWidth)
            );
        };

        this.getElementBounds = function getElementBounds(selector) {
            try {
                var clipRect = this.findOne(selector).getBoundingClientRect();
                return {
                    top: clipRect.top,
                    left: clipRect.left,
                    width: clipRect.width,
                    height: clipRect.height
                };
            } catch (e) {
                this.log("Unable to fetch bounds for element " + selector, "warning");
            }
        };

        this.getElementsBounds = function getElementsBounds(selector) {
            var elements = this.findAll(selector);
            try {
                return Array.prototype.map.call(elements, function(element) {
                    var clipRect = element.getBoundingClientRect();
                    return {
                        top: clipRect.top,
                        left: clipRect.left,
                        width: clipRect.width,
                        height: clipRect.height
                    };
                });
            } catch (e) {
                this.log("Unable to fetch bounds for elements matching " + selector, "warning");
            }
        };
        this.getElementInfo = function getElementInfo(selector) {
            var element = this.findOne(selector);
            var bounds = this.getElementBounds(selector);
            var attributes = {};
            [].forEach.call(element.attributes, function(attr) {
                attributes[attr.name.toLowerCase()] = attr.value;
            });
            return {
                nodeName: element.nodeName.toLowerCase(),
                attributes: attributes,
                tag: element.outerHTML,
                html: element.innerHTML,
                text: element.textContent || element.innerText,
                x: bounds.left,
                y: bounds.top,
                width: bounds.width,
                height: bounds.height,
                visible: this.visible(selector)
            };
        };

        this.getElementsInfo = function getElementsInfo(selector) {
            var bounds = this.getElementsBounds(selector);
            var eleVisible = this.elementVisible;
            return [].map.call(this.findAll(selector), function(element, index) {
                var attributes = {};
                [].forEach.call(element.attributes, function(attr) {
                    attributes[attr.name.toLowerCase()] = attr.value;
                });
                return {
                    nodeName: element.nodeName.toLowerCase(),
                    attributes: attributes,
                    tag: element.outerHTML,
                    html: element.innerHTML,
                    text: element.textContent || element.innerText,
                    x: bounds[index].left,
                    y: bounds[index].top,
                    width: bounds[index].width,
                    height: bounds[index].height,
                    visible: eleVisible(element)
                };
            });
        };

        this.getElementByXPath = function getElementByXPath(expression, scope) {
            scope = scope || this.options.scope;
            var a = document.evaluate(expression, scope, this.xpathNamespaceResolver, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
            if (a.snapshotLength > 0) {
                return a.snapshotItem(0);
            }
        };

        this.getElementsByXPath = function getElementsByXPath(expression, scope) {
            scope = scope || this.options.scope;
            var nodes = [];
            var a = document.evaluate(expression, scope, this.xpathNamespaceResolver, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
            for (var i = 0; i < a.snapshotLength; i++) {
                nodes.push(a.snapshotItem(i));
            }
            return nodes;
        };

        this.xpathNamespaceResolver = function xpathNamespaceResolver(prefix) {
          return XPATH_NAMESPACE[prefix] || null;
        };

        this.getFieldValue = function getFieldValue(selector, scope) {
            var self = this;
            var fields = this.findAll(selector, scope);
            var type;

            if (!(fields instanceof NodeList || fields instanceof Array)) {
                this.log("attempting to fetch field element from selector: '" + selector + "'", "info");
                fields = this.findAll('[name="' + selector + '"]');
            }

            if (fields && fields.length > 1) {
                type = fields[0].hasAttribute('type') ? fields[0].getAttribute('type') : "other";
                fields = [].filter.call(fields, function(elm){
                    if (elm.nodeName.toLowerCase() === 'input' &&
                        ['checkbox', 'radio'].indexOf(elm.getAttribute('type')) !== -1) {
                        return elm.checked;
                    }
                    return true;
                });
            }

            if (fields.length === 0 ) {
                return type !== "radio" ? [] : undefined;
            }

            if (fields.length > 1 ) {
                return [].map.call(fields, function(elm) {
                    var ret = self.getField(elm);
                    return ret && type === 'checkbox' ? elm.value : ret;
                });
            }

            return this.getField(fields[0]);
        };

        this.getField = function getField(field) {
            var nodeName, type;

            if (!(field instanceof HTMLElement)) {
                var error = new Error('getFieldValue: Invalid field ; only HTMLElement is supported');
                error.name = 'FieldNotFound';
                throw error;
            }

            nodeName = field.nodeName.toLowerCase();
            type = field.hasAttribute('type') ? field.getAttribute('type').toLowerCase() : 'text';
            if (nodeName === "select" && field.multiple) {
                return [].filter.call(field.options, function(option){
                    return !!option.selected;
                }).map(function(option){
                    return option.value || option.text;
                });
            }
            if (type === 'radio') {
                return field.checked ? field.value : null;
            }
            if (type === 'checkbox') {
                return field.checked;
            }
            return field.value || '';
        };

        this.getFormValues = function getFormValues(form) {
            var self = this;
            var values = {}, checked = {};

            if (!(form instanceof HTMLElement) || typeof form === "string") {
                this.log("attempting to fetch form element from selector: '" + form + "'", "info");
                try {
                    form = this.findOne(form);
                } catch (e) {
                    this.log("invalid form selector provided: '" + form + "'");
                    return {};
                }
            }

            [].forEach.call(form.elements, function(elm) {
                var name = elm.getAttribute('name');
                var value = self.getField(elm);
                var multi = !!value && elm.hasAttribute('type') &&
                            elm.type === 'checkbox' ? elm.value : value;
                if (!!name && value !== null && !(elm.type === 'checkbox' && value === false)) {
                    if (typeof values[name] === "undefined") {
                        values[name] = value;
                        checked[name] = multi;
                    } else {
                        if (!Array.isArray(values[name])) {
                            values[name] = [checked[name]];
                        }
                        values[name].push(multi);
                    }
                }
            });
            return values;
        };
        this.log = function log(message, level) {
            console.log("[casper:" + (level || "debug") + "] " + message);
        };

        this.makeSelector = function makeSelector(selector, type){
            type = type || 'xpath';
            var ret;

            if (typeof selector === "object") { 
                selector = selector.path;
            }

            switch (type) {
                case 'css':
                    ret = selector;
                    break;
                case 'name': 
                case 'names':
                    ret = '[name="' + selector + '"]';
                    break;
                case 'label': 
                case 'labels':
                    ret = {type: 'xpath', path: '//*[@id=string(//label[text()="' + selector + '"]/@for)]'};
                    break;
                case 'xpath':
                    ret = {type: 'xpath', path: selector};
                    break;
                default:
                    throw new Error("Unsupported selector type: " + type);
            }

            return ret;
        };

        this.mouseEvent = function mouseEvent(type, selector, x, y) {
            var elem = this.findOne(selector);
            if (!elem || ( !this.elementVisible(elem) && elem.nodeName.toUpperCase() !== "AREA")) {
                this.log("mouseEvent(): Couldn't find any element matching '" +
                    selector + "' selector", "error");
                return false;
            }

            var convertNumberToIntAndPercentToFloat = function (a, def){
                return !!a && !isNaN(a) && parseInt(a, 10) ||
                    !!a && !isNaN(parseFloat(a)) && parseFloat(a) >= 0 &&
                    parseFloat(a) <= 100 && parseFloat(a) / 100 ||
                def;
            };
            try {
                var evt = document.createEvent("MouseEvents");
                var px = convertNumberToIntAndPercentToFloat(x, 0.5),
                    py = convertNumberToIntAndPercentToFloat(y, 0.5);
                try {
                    var bounds = elem.getBoundingClientRect();
                    px = Math.floor(bounds.width  * (px - (px ^ 0)).toFixed(10)) + (px ^ 0) + bounds.left;
                    py = Math.floor(bounds.height * (py - (py ^ 0)).toFixed(10)) + (py ^ 0) + bounds.top;
                } catch (e) {
                    px = 1; py = 1;
                }
                evt.initMouseEvent(type, true, true, window, 1, 1, 1, px, py, false, false, false, false,
                    type !== "contextmenu" ? 0 : 2, elem);
                elem.dispatchEvent(evt);
                return true;
            } catch (e) {
                this.log("Failed dispatching " + type + "mouse event on " + selector + ": " + e, "error");
                return false;
            }
        };

        this.processSelector = function processSelector(selector) {
            var selectorObject = {
                toString: function toString() {
                    return this.type + ' selector: ' + this.path;
                }
            };
            if (typeof selector === "string") {
                // defaults to CSS selector
                selectorObject.type = "css";
                selectorObject.path = selector;
                return selectorObject;
            } else if (typeof selector === "object") {
                // validation
                if (!selector.hasOwnProperty('type') || !selector.hasOwnProperty('path')) {
                    throw new Error("Incomplete selector object");
                } else if (SUPPORTED_SELECTOR_TYPES.indexOf(selector.type) === -1) {
                    throw new Error("Unsupported selector type: " + selector.type);
                }
                if (!selector.hasOwnProperty('toString')) {
                    selector.toString = selectorObject.toString;
                }
                return selector;
            }
            throw new Error("Unsupported selector type: " + typeof selector);
        };

        this.removeElementsByXPath = function removeElementsByXPath(expression) {
            var a = document.evaluate(expression, document, null,
                XPathResult.UNORDERED_NODE_SNAPSHOT_TYPE, null);
            for (var i = 0; i < a.snapshotLength; i++) {
                a.snapshotItem(i).parentNode.removeChild(a.snapshotItem(i));
            }
        };
        this.scrollTo = function scrollTo(x, y) {
            window.scrollTo(parseInt(x || 0, 10), parseInt(y || 0, 10));
        };

        this.scrollToBottom = function scrollToBottom() {
            this.scrollTo(0, this.getDocumentHeight());
        };
        this.sendAJAX = function sendAJAX(url, method, data, async, settings) {
            var xhr = new XMLHttpRequest(),
                dataString = "",
                dataList = [];
            var CONTENT_TYPE_HEADER = "Content-Type";
            method = method && method.toUpperCase() || "GET";
            var contentTypeValue = settings && settings.contentType || "application/x-www-form-urlencoded";
            xhr.open(method, url, !!async);
            this.log("sendAJAX(): Using HTTP method: '" + method + "'", "debug");
            if (settings && settings.overrideMimeType) {
                xhr.overrideMimeType(settings.overrideMimeType);
            }
            if (settings && settings.headers) {
               for (var header in settings.headers) {
                   if (header === CONTENT_TYPE_HEADER) {
                       contentTypeValue = settings.headers[header];
                   } else {
                       xhr.setRequestHeader(header, settings.headers[header]);
                   }
              }
            }
            if (method === "POST") {
                if (typeof data === "object") {
                    for (var k in data) {
                        if (data.hasOwnProperty(k)) {
                            dataList.push(form_urlencoded(k) + "=" +
                             form_urlencoded(data[k].toString()));
                        }
                    }
                    dataString = dataList.join('&');
                    this.log("sendAJAX(): Using request data: '" + dataString + "'", "debug");
                } else if (typeof data === "string") {
                    dataString = data;
                }
                xhr.setRequestHeader(CONTENT_TYPE_HEADER, contentTypeValue);
            }
            xhr.send(method === "POST" ? dataString : null);
            return xhr.responseText;
        };
        this.setFieldValue = function setFieldValue(selector, value, scope) {
            var self = this;
            var fields = this.findAll(selector, scope);
            var values = value;

            if (!Array.isArray(value)) {
                values = [value];
            }

            if (fields && fields.length > 1) {
                fields = [].filter.call(fields, function(elm){
                    if (elm.nodeName.toLowerCase() === 'input' &&
                        ['checkbox', 'radio'].indexOf(elm.getAttribute('type')) !== -1) {
                        return values.indexOf(elm.getAttribute('value')) !== -1;
                    }
                    return true;
                });
                [].forEach.call(fields, function(elm) {
                    self.setField(elm, value);
                });
            } else {
                this.setField(fields[0], value);
            }
            return true;
        };
        this.setField = function setField(field, value) {
            var logValue, out, filter;
            value = logValue = value || "";

            if (!(field instanceof HTMLElement)) {
                var error = new Error('setField: Invalid field ; only HTMLElement is supported');
                error.name = 'FieldNotFound';
                throw error;
            }

            if (this.options && this.options.safeLogs && field.getAttribute('type') === "password") {
                logValue = new Array(('' + value).length + 1).join("*");
            }
            this.log('Set "' + field.getAttribute('name') + '" field value to ' + logValue, "debug");

            try {
                field.focus();
            } catch (e) {
                this.log("Unable to focus() input field " + field.getAttribute('name') + ": " + e, "warning");
            }

            filter = String(field.getAttribute('type') ? field.getAttribute('type') : field.nodeName).toLowerCase();
            switch (filter) {
                case "checkbox":
                case "radio":
                    field.checked = value ? true : false;
                    break;
                case "file":
                    throw {
                        name: "FileUploadError",
                        message: "File field must be filled using page.uploadFile",
                        path: value,
                        id: field.id || null
                    };
                    break;
                case "select":
                    if (field.multiple) {
                        [].forEach.call(field.options, function(option) {
                            option.selected = value.indexOf(option.value) !== -1;
                        });
                        if (field.value === "") {
                            [].forEach.call(field.options, function(option) {
                                option.selected = value.indexOf(option.text) !== -1;
                            });
                        }
                    } else {
                        if (value === "") {
                            field.selectedIndex = -1;
                        } else {
                            field.value = value;
                        }
                        if (field.value !== value) {
                            [].some.call(field.options, function(option) {
                                option.selected = value === option.text;
                                return value === option.text;
                            });
                        }
                    }
                    break;
                default:
                    field.value = value;
            }

            ['change', 'input'].forEach(function(name) {
                var event = document.createEvent("HTMLEvents");
                event.initEvent(name, true, true);
                field.dispatchEvent(event);
            });
            try {
                field.blur();
            } catch (err) {
                this.log("Unable to blur() input field " + field.getAttribute('name') +
                 ": " + err, "warning");
            }
            return out;
        };

        this.setScope = function setScope(selector) {
            var scope = !(this.options.scope instanceof HTMLElement) ? this.getCssSelector(this.options.scope) : "";
            this.options.scope = (selector !== "") ? this.findOne(selector) : document;
            return scope;
        };

        this.visible = function visible(selector) {
            return [].some.call(this.findAll(selector), this.elementVisible);
        };

        this.allVisible = function allVisible(selector) {
            return [].every.call(this.findAll(selector), this.elementVisible);
        };
    };
})(typeof exports ===  "object" && !(exports instanceof Element) ? exports : window);
