require = patchRequire(require);
var utils = require('utils');
var f = utils.format;

function create() {
    "use strict";
    return new Stack();
}
exports.create = create;

var Stack = function Stack(){};
exports.Stack = Stack;

Stack.prototype = [];

Stack.prototype.clean = function clean() {
    "use strict";
    var self = this;

    this.forEach(function(popup, index) {
        if (popup.parent === null || typeof popup.parent === "undefined") {
            self.splice(index, 1);
        }
    });
    return this.length;
};

Stack.prototype.find = function find(popupInfo) {
    "use strict";
    var popup, type = utils.betterTypeOf(popupInfo);
    switch (type) {
        case "regexp":
            popup = this.findByRegExp(popupInfo);
            break;
        case "string":
            popup = this.findByURL(popupInfo);
            break;
        case "object":
            popup = this.findByUrlNameTitle(popupInfo);
            break;
        case "number":
            popup = this.findByIndex(popupInfo);
            break;
        case "qtruntimeobject": 
            popup = popupInfo;
            if (!utils.isWebPage(popup) || !this.some(function(popupPage) {
                if (popupInfo.id && popupPage.id) {
                    return popupPage.id === popup.id;
                }
                return popupPage.url === popup.url;
            })) {
                throw new CasperError("Invalid or missing popup.");
            }
            break;
        default:
            throw new CasperError(f("Invalid popupInfo type: %s.", type));
    }
    return popup;
};

Stack.prototype.findByRegExp = function findByRegExp(regexp) {
    "use strict";
    var popup = this.filter(function(popupPage) {
        return regexp.test(popupPage.url);
    })[0];
    if (!popup) {
        throw new CasperError(f("Couldn't find popup with url matching pattern %s", regexp));
    }
    return popup;
};
Stack.prototype.findByTitle = function findByTitle(string) {
    "use strict";
    var popup = this.filter(function(popupPage) {
        return popupPage.title.indexOf(string) !== -1;
    })[0];
    if (!popup) {
        throw new CasperError(f("Couldn't find popup with title containing '%s'", string));
    }
    return popup;
};

Stack.prototype.findByURL = function findByURL(string) {
    "use strict";
    var popup = this.filter(function(popupPage) {
        return popupPage.url.indexOf(string) !== -1;
    })[0];
    if (!popup) {
        throw new CasperError(f("Couldn't find popup with url containing '%s'", string));
    }
    return popup;
};

Stack.prototype.findByUrlNameTitle = function findByUrlNameTitle(object) {
    "use strict";
    var popup = null;
    try {
        if (typeof object.url !== "undefined") {
            popup = this.findByUrl(object.url);
        }
        if (!popup && typeof object.title !== "undefined") {
            popup = this.findByTitle(object.title);
        }
        if (!popup && typeof object.windowName !== "undefined") {
            popup = this.findByWindowName(object.windowName);
        }
    } catch(e){}

    if (!popup) {
        throw new CasperError(f("Couldn't find popup with object '%s'", JSON.stringify(object)));
    }
    return popup;
};

Stack.prototype.findByWindowName = function findByWindowName(string) {
    "use strict";
    
    var popup = this.filter(function(popupPage) {
        return popupPage.windowName.indexOf(string) !== -1 || popupPage.windowNameBackUp.indexOf(string) !== -1;
    })[0];
    if (!popup) {
        throw new CasperError(f("Couldn't find popup with name containing '%s'", string));
    }
    return popup;
};

Stack.prototype.findByIndex = function findByIndex(index) {
    "use strict";
    var popup = this[index];
    if (!popup) {
        throw new CasperError(f("Couldn't find popup with index containing '%d'", index));
    }
    return popup;
};

Stack.prototype.list = function list() {
    "use strict";
    return this.map(function(popup) {
        try {
            return popup.url;
        } catch (e) {
            return '<deleted>';
        }
    });
};

Stack.prototype.toString = function toString() {
    "use strict";
    return f("[Object Stack], having %d popup(s)" % this.length);
};
