require = patchRequire(require);
var utils = require('utils');
var fs = require('fs');
var TestSuiteResult = require('tester').TestSuiteResult;

function generateClassName(classname) {
    "use strict";
    classname = (classname || "").replace(phantom.casperPath, "").trim();
    var script = classname || phantom.casperScript || "";
    if (script.indexOf(fs.workingDirectory) === 0) {
        script = script.substring(fs.workingDirectory.length + 1);
    }
    if (script.indexOf('/') === 0) {
        script = script.substring(1, script.length);
    }
    if (~script.indexOf('.')) {
        script = script.substring(0, script.lastIndexOf('.'));
    }
    if (!script && phantom.casperScript) {
        script = phantom.casperScript;
    }
    return script || "unknown";
}

exports.create = function create() {
    "use strict";
    return new XUnitExporter();
};

function XUnitExporter() {
    "use strict";

    this.setupDocument();

    this.results = undefined;
    this.rendered = false;
}
exports.XUnitExporter = XUnitExporter;
XUnitExporter.prototype.getXML = function getXML() {
    "use strict";

    var self = this;

    if (!(this.results instanceof TestSuiteResult)) {
        throw new CasperError('Results not set, cannot get XML.');
    }
    this.results.forEach(function(result) {
        var suiteNode = utils.node('testsuite', {
            name: result.name,
            tests: result.assertions,
            failures: result.failed,
            errors: result.crashed,
            time: utils.ms2seconds(result.calculateDuration()),
            timestamp: (new Date()).toISOString(),
            'package': generateClassName(result.file)
        });
        result.passes.forEach(function(success) {
            var testCase = utils.node('testcase', {
                name: success.message || success.standard,
                classname: generateClassName(success.file),
                time: utils.ms2seconds(~~success.time)
            });
            suiteNode.appendChild(testCase);
        });
        result.failures.forEach(function(failure) {
            var testCase = utils.node('testcase', {
                name: failure.name || failure.message || failure.standard,
                classname: generateClassName(failure.file),
                time: utils.ms2seconds(~~failure.time)
            });
            var failureNode = utils.node('failure', {
                type: failure.type || "failure"
            });
            failureNode.appendChild(self._xmlDocument.createCDATASection(failure.message || "no message left"));
            if (failure.values && failure.values.error instanceof Error) {
                var errorNode = utils.node('error', {
                    type: utils.betterTypeOf(failure.values.error)
                });
                errorNode.appendChild(self._xmlDocument.createCDATASection(failure.values.error.stack));
                testCase.appendChild(errorNode);
            }
            testCase.appendChild(failureNode);
            suiteNode.appendChild(testCase);
        });
        result.errors.forEach(function(error) {
            var errorNode = utils.node('error', {
                type: error.name
            });
            errorNode.appendChild(self._xmlDocument.createCDATASection(error.stack ? error.stack : error.message));
            suiteNode.appendChild(errorNode);
        });
        var warningNode = utils.node('system-out');
        warningNode.appendChild(self._xmlDocument.createCDATASection(result.warnings.join('\n')));
        suiteNode.appendChild(warningNode);
        this._xml.appendChild(suiteNode);
    }.bind(this));

    this._xml.setAttribute('time', utils.ms2seconds(this.results.calculateDuration()));

    this.rendered = true;

    return this._xmlDocument;
};

XUnitExporter.prototype.getSerializedXML = function getSerializedXML() {
    "use strict";
    var serializer = new XMLSerializer(),
        document;

    if ( !this.rendered ) {
        document = this.getXML();
    }
    return '<?xml version="1.0" encoding="UTF-8"?>' + serializer.serializeToString(document);
};

XUnitExporter.prototype.setResults = function setResults(results) {
    "use strict";
    if (!(results instanceof TestSuiteResult)) {
        throw new CasperError('Invalid results type.');
    }
    this.results = results;

    this.setupDocument();
    this.rendered = false;

    return results;
};

XUnitExporter.prototype.setupDocument = function() {
    this._xmlDocument = document.implementation.createDocument("", "");
    this._xml = this._xmlDocument.appendChild(this._xmlDocument.createElement("testsuites"));
};
